import json
from pathlib import Path
from typing import Any, List, Mapping, Optional

import jsonschema
import typer
from apischema.json_schema import JsonSchemaVersion, deserialization_schema
from ruamel.yaml import YAML

from ibek import __version__

from .gen_scripts import create_boot_script, create_db_script, ioc_deserialize
from .ioc import IOC, make_entity_classes
from .support import Support

cli = typer.Typer()
yaml = YAML()


def make_schema(cls: type) -> Mapping[str, Any]:
    return deserialization_schema(cls, version=JsonSchemaVersion.DRAFT_7)


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@cli.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print the version of ibek and exit",
    )
):
    """Do 3 things..."""


@cli.command()
def ibek_schema(
    output: Path = typer.Argument(..., help="The filename to write the schema to")
):
    """Produce JSON global schema for all <support_module>.ibek.support.yaml files"""
    schema = json.dumps(make_schema(Support), indent=2)
    output.write_text(schema)


@cli.command()
def ioc_schema(
    definitions: List[Path] = typer.Argument(
        ..., help="The filepath to a support module definition file"
    ),
    output: Path = typer.Argument(..., help="The filename to write the schema to"),
    no_schema: bool = typer.Option(False, help="disable schema checking"),
):
    """
    Create a json schema from a <support_module>.ibek.support.yaml file
    """

    # first check the definition file with jsonschema since it has more
    # legible error messages than apischema
    for definition in definitions:
        support_dict = YAML(typ="safe").load(definition)
        if not no_schema:
            schema_support = make_schema(Support)
            jsonschema.validate(support_dict, schema_support)

        support = Support.deserialize(support_dict)
        make_entity_classes(support)

    schema = json.dumps(make_schema(IOC), indent=2)
    output.write_text(schema)


@cli.command()
def build_startup(
    instance: Path = typer.Argument(
        ..., help="The filepath to the ioc instance entity file"
    ),
    definitions: List[Path] = typer.Argument(
        ..., help="The filepath to a support module definition file"
    ),
    out: Path = typer.Option(
        default="config/st.cmd",
        help="Path to output startup script",
    ),
    db_out: Path = typer.Option(
        default="config/make_db.sh",
        help="Path to output database expansion shell script",
    ),
):
    """
    Build a startup script for an IOC instance
    """

    ioc_instance = ioc_deserialize(instance, definitions)

    script_txt = create_boot_script(ioc_instance)

    out.parent.mkdir(parents=True, exist_ok=True)

    with out.open("w") as stream:
        stream.write(script_txt)

    db_txt = create_db_script(ioc_instance)

    with db_out.open("w") as stream:
        stream.write(db_txt)


# test with:
#     pipenv run python -m ibek
if __name__ == "__main__":
    cli()



"""
A few global defintions
"""
from typing import TypeVar

from apischema import schema

#: A generic Type for use in type hints
T = TypeVar("T")


def desc(description: str):
    """a description Annotation to add to our Entity derived Types"""
    return schema(description=description)



"""
Functions for generating an IocInstance derived class from a
support module definition YAML file
"""
from __future__ import annotations

import builtins
import types
from dataclasses import Field, dataclass, field, make_dataclass
from typing import Any, Dict, List, Mapping, Sequence, Tuple, Type, cast

from apischema import (
    Undefined,
    ValidationError,
    cache,
    deserialize,
    deserializer,
    identity,
    schema,
)
from apischema.conversions import Conversion, reset_deserializers
from apischema.metadata import conversion
from jinja2 import Template
from typing_extensions import Annotated as A
from typing_extensions import Literal

from . import modules
from .globals import T, desc
from .support import Definition, IdArg, ObjectArg, Support
from .utils import UTILS


class Entity:
    """
    A baseclass for all generated Entity classes. Provides the
    deserialize entry point.
    """

    # a link back to the Definition Object that generated this Definition
    __definition__: Definition

    entity_enabled: bool

    def __post_init__(self: "Entity"):
        # If there is an argument which is an id then allow deserialization by that
        args = self.__definition__.args
        ids = set(a.name for a in args if isinstance(a, IdArg))
        assert len(ids) <= 1, f"Multiple id args {list(ids)} defined in {args}"
        if ids:
            # A string id, use that
            inst_id = getattr(self, ids.pop())
            assert inst_id not in id_to_entity, f"Already got an instance {inst_id}"
            id_to_entity[inst_id] = self

            # TODO - not working as printing own ID
            setattr(self, "__str__", inst_id)

        # add in the global __utils__ object for state sharing
        self.__utils__ = UTILS

        # copy 'values' from the definition into the Entity
        for value in self.__definition__.values:
            setattr(self, value.name, value.value)

        # Jinja expansion of any string args/values in the Entity's attributes
        for arg, value in self.__dict__.items():
            if isinstance(value, str):
                jinja_template = Template(value)
                rendered = jinja_template.render(self.__dict__)
                setattr(self, arg, rendered)


id_to_entity: Dict[str, Entity] = {}


def make_entity_class(definition: Definition, support: Support) -> Type[Entity]:
    """
    We can get a set of Definitions by deserializing an ibek
    support module definition YAML file.

    This function then creates an Entity derived class from each Definition.

    See :ref:`entities`
    """
    fields: List[Tuple[str, type, Field[Any]]] = []

    # add in each of the arguments
    for arg in definition.args:
        # make_dataclass can cope with string types, so cast them here rather
        # than lookup
        metadata: Any = None
        arg_type: Type
        if isinstance(arg, ObjectArg):

            def lookup_instance(id):
                try:
                    return id_to_entity[id]
                except KeyError:
                    raise ValidationError(f"{id} is not in {list(id_to_entity)}")

            metadata = conversion(
                deserialization=Conversion(lookup_instance, str, Entity)
            ) | schema(extra={"vscode_ibek_plugin_type": "type_object"})
            arg_type = Entity
        elif isinstance(arg, IdArg):
            arg_type = str
            metadata = schema(extra={"vscode_ibek_plugin_type": "type_id"})
        else:
            # arg.type is str, int, float, etc.
            arg_type = getattr(builtins, arg.type)
        if arg.description:
            arg_type = A[arg_type, desc(arg.description)]  # type: ignore
        if arg.default is Undefined:
            fld = field(metadata=metadata)
        else:
            fld = field(metadata=metadata, default=arg.default)
        fields.append((arg.name, arg_type, fld))  # type: ignore

    # put the literal name in as 'type' for this Entity this gives us
    # a unique key for each of the entity types we may instantiate
    full_name = f"{support.module}.{definition.name}"
    fields.append(
        (
            "type",
            Literal[full_name],  # type: ignore
            field(default=cast(Any, full_name)),
        )
    )

    # add a field so we can control rendering of the entity without having to delete
    # it
    fields.append(("entity_enabled", bool, field(default=cast(Any, True))))

    namespace = dict(__definition__=definition)

    # make the Entity derived dataclass for this EntityClass, with a reference
    # to the Definition that created it
    entity_cls = make_dataclass(full_name, fields, bases=(Entity,), namespace=namespace)
    deserializer(Conversion(identity, source=entity_cls, target=Entity))
    return entity_cls


def make_entity_classes(support: Support) -> types.SimpleNamespace:
    """Create `Entity` subclasses for all `Definition` objects in the given
    `Support` instance, put them in a namespace in `ibek.modules` and return
    it"""
    module = types.SimpleNamespace()
    assert not hasattr(
        modules, support.module
    ), f"Entity classes already created for {support.module}"
    setattr(modules, support.module, module)
    modules.__all__.append(support.module)
    for definition in support.defs:
        entity_cls = make_entity_class(definition, support)
        setattr(module, definition.name, entity_cls)
    return module


def clear_entity_classes():
    """Reset the modules namespaces, deserializers and caches of defined Entity
    subclasses"""
    id_to_entity.clear()
    while modules.__all__:
        delattr(modules, modules.__all__.pop())
    reset_deserializers(Entity)
    cache.reset()


@dataclass
class IOC:
    """
    Used to load an IOC instance entities yaml file into memory using
    IOC.deserialize(YAML().load(ioc_instance_yaml)).

    Before loading the entities file all Entity classes that it contains
    must be defined in modules.py. This is achieved by deserializing all
    support module definitions yaml files used by this IOC and calling
    make_entity_classes(support_module).
    """

    ioc_name: A[str, desc("Name of IOC instance")]
    description: A[str, desc("Description of what the IOC does")]
    entities: A[Sequence[Entity], desc("List of entities this IOC instantiates")]
    generic_ioc_image: A[str, desc("The generic IOC container image registry URL")]

    @classmethod
    def deserialize(cls: Type[T], d: Mapping[str, Any]) -> T:
        return deserialize(cls, d)



from pathlib import Path

from pytest import fixture
from ruamel.yaml import YAML

from ibek.ioc import clear_entity_classes, make_entity_classes
from ibek.support import Support


def get_support(samples: Path, yaml_file: str) -> Support:
    """
    Get a support object from the sample YAML directory
    """
    # load from file
    d = YAML(typ="safe").load(samples / f"{yaml_file}")
    # create a support object from that dict
    support = Support.deserialize(d)
    return support


@fixture
def ibek_defs():
    return Path(__file__).parent.parent / "ibek-defs"


@fixture
def samples():
    return Path(__file__).parent / "samples"


@fixture
def pmac_support(ibek_defs):
    return get_support(ibek_defs / "pmac", "pmac.ibek.support.yaml")


@fixture
def pmac_classes(pmac_support):
    # clear the entity classes to make sure there's nothing left
    clear_entity_classes()

    # make entity subclasses for everything defined in it
    namespace = make_entity_classes(pmac_support)

    # return the namespace so that callers have access to all of the
    # generated dataclasses
    return namespace


@fixture
def epics_support(ibek_defs):
    return get_support(ibek_defs / "_global", "epics.ibek.support.yaml")


@fixture
def epics_classes(epics_support):
    # clear the entity classes to make sure there's nothing left
    clear_entity_classes()

    # make entity subclasses for everything defined in it
    namespace = make_entity_classes(epics_support)

    # return the namespace so that callers have access to all of the
    # generated dataclasses
    return namespace



"""
The Support Class represents a deserialized <MODULE_NAME>.ibek.support.yaml file.
It contains a hierarchy of Entity dataclasses.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Mapping, Optional, Sequence, Type, Union

from apischema import Undefined, UndefinedType, deserialize, deserializer, identity
from apischema.conversions import Conversion
from typing_extensions import Annotated as A
from typing_extensions import Literal

from .globals import T, desc


@dataclass
class Arg:
    """Base class for all Argument Types"""

    name: A[str, desc("Name of the argument that the IOC instance should pass")]
    description: A[str, desc("Description of what the argument will be used for")]
    type: str
    default: Any

    # https://wyfo.github.io/apischema/latest/examples/subclass_union/
    def __init_subclass__(cls):
        # Deserializers stack directly as a Union
        deserializer(Conversion(identity, source=cls, target=Arg))


Default = A[
    Union[Optional[T], UndefinedType],
    desc("If given, and instance doesn't supply argument, what value should be used"),
]


# FloatArg must be defined before StrArg, otherwise we get:
#
#    TypeError: Invalid JSON type <class 'ruamel.yaml.scalarfloat.ScalarFloat'>
#
# During Support.deserialize, when default float values in pmac.ibek.support.yaml do not
# have a trailing 'f'. It is due to the order of declaration of subclasses of
# Arg. When StrArg is before FloatArg, apischema attempts to deserialize as a
# string first. The coercion from str to number requires a trailing f if there
# is a decimal.
@dataclass
class FloatArg(Arg):
    """An argument with a float value"""

    type: Literal["float"] = "float"
    default: Default[float] = Undefined


@dataclass
class StrArg(Arg):
    """An argument with a str value"""

    type: Literal["str"] = "str"
    default: Default[str] = Undefined


@dataclass
class IntArg(Arg):
    """An argument with an int value"""

    type: Literal["int"] = "int"
    default: Default[int] = Undefined


@dataclass
class BoolArg(Arg):
    """An argument with an bool value"""

    type: Literal["bool"] = "bool"
    default: Default[bool] = Undefined


@dataclass
class ObjectArg(Arg):
    """A reference to another entity defined in this IOC"""

    type: Literal["object"] = "object"
    default: Default[str] = Undefined


@dataclass
class IdArg(Arg):
    """Explicit ID argument that an object can refer to"""

    type: Literal["id"] = "id"
    default: Default[str] = Undefined


@dataclass
class Database:
    """
    A database file that should be loaded by the startup script and its args
    """

    file: A[str, desc("Filename of the database template in <module_root>/db")]
    include_args: A[
        Sequence[str], desc("List of args to pass through to database")
    ] = ()
    define_args: A[str, desc("Arg string list to be generated as Jinja template")] = ""


@dataclass
class EnvironmentVariable:
    """
    An environment variable that should be set in the startup script
    """

    name: A[str, desc("Name of environment variable")]
    value: A[str, desc("Value to set")]


@dataclass
class Function:
    """
    A script snippet that defines a function to call
    """

    name: A[str, desc("Name of the function to call")]
    args: A[Dict[str, Any], desc("The arguments IOC instance should supply")]
    header: A[str, desc("commands/comments to appear before the function")] = ""
    once: A[bool, desc("If true, only call the function once")] = False
    type: Literal["function"] = "function"


@dataclass
class Once:
    """
    A script snippet that should for the first occurrence only
    """

    type: Literal["once"] = "once"
    value: A[str, desc("Startup script snippets defined as Jinja template")] = ""


@dataclass
class Value:
    """A calculated string value for a definition"""

    name: A[str, desc("Name of the value that the IOC instance will expose")]
    description: A[str, desc("Description of what the value will be used for")]
    value: A[str, desc("The contents of the value")]

    def __str__(self):
        return self.value


@dataclass
class Definition:
    """
    A single definition of a class of Entity that an IOC instance may instantiate
    """

    name: A[str, desc("Publish Definition as type <module>.<name> for IOC instances")]
    description: A[str, desc("Describes the purpose of the definition")]
    args: A[Sequence[Arg], desc("The arguments IOC instance should supply")] = ()
    values: A[Sequence[Value], desc("The values IOC instance should supply")] = ()
    databases: A[Sequence[Database], desc("Databases to instantiate")] = ()
    script: A[
        Sequence[Union[str, Function, Once]],
        desc("Startup script snippets defined as Jinja template or function"),
    ] = ()
    env_vars: A[
        Sequence[EnvironmentVariable],
        desc("Environment variables to set in the boot script"),
    ] = ()
    post_ioc_init: A[
        Sequence[str], desc("Entries to add post iocInit(), such as dbpf")
    ] = ()


@dataclass
class Support:
    """
    Lists the definitions for a support module, this defines what Entities it supports

    Provides the deserialize entry point.
    """

    module: A[str, desc("Support module name, normally the repo name")]
    defs: A[
        Sequence[Definition],
        desc("The definitions an IOC can create using this module"),
    ]

    @classmethod
    def deserialize(cls: Type[T], d: Mapping[str, Any]) -> T:
        return deserialize(cls, d)



"""
Functions for rendering lines in the boot script using Jinja2
"""

from typing import Callable, List, Optional, Union

from jinja2 import Template

from .ioc import IOC, Entity
from .support import Function, Once


class Render:
    """
    A class for generating snippets of startup script / EPICS DB
    by using Jinja to combine snippet templates from support module
    definition yaml with substitution values supplied in ioc entity yaml
    """

    def __init__(self: "Render"):
        self.once_done: List[str] = []

    def render_text(self, instance: Entity, text: str, once=False, suffix="") -> str:
        """
        Add in the next line of text, honouring the ``once`` flag which will
        only add the line once per IOC.

        Jinja rendering of values/args has already been done in Entity.__post_init__
        but we pass all strings though jinja again to render any other jinja
        in the IOC (e.g. database and function entries)

        ``once`` uses the name of the definition + suffix to track which lines
        have been rendered already. The suffix can be used where a given
        Entity has more than one element to render once (e.g. functions)
        """

        if once:
            name = instance.__definition__.name + suffix
            if name not in self.once_done:
                self.once_done.append(name)
            else:
                return ""

        # Render Jinja entries in the text
        jinja_template = Template(text)
        result = jinja_template.render(instance.__dict__)  # type: ignore

        if result == "":
            return ""

        return result + "\n"

    def render_function(self, instance: Entity, function: Function) -> str:
        """
        render a Function object that represents a function call in the IOC
        startup script
        """

        # initial function comment appears after newline for prettier formatting
        comment = f"\n# {function.name} "
        call = f"{function.name} "
        for name, value in function.args.items():
            comment += f"{name} "
            call += f"{value} "

        text = (
            self.render_text(instance, comment.strip(" "), once=True, suffix="func")
            + self.render_text(instance, function.header, once=True, suffix="func_hdr")
            + self.render_text(instance, call.strip(" "))
        )

        return text

    def render_script(self, instance: Entity) -> Optional[str]:
        """
        render the startup script by combining the jinja template from
        an entity with the arguments from an Entity
        """

        script = ""
        script_items = getattr(instance.__definition__, "script")
        for line in script_items:
            if isinstance(line, str):
                script += self.render_text(instance, line)
            elif isinstance(line, Once):
                script += self.render_text(instance, line.value, True)
            elif isinstance(line, Function):
                script += self.render_function(instance, line)

        return script

    def render_database(self, instance: Entity) -> Optional[str]:
        """
        render the lines required to instantiate database by combining the
        templates from the Entity's database list with the arguments from
        an Entity
        """
        templates = instance.__definition__.databases
        # The entity may not instantiate any database templates
        if not templates:
            return None
        jinja_txt = ""

        for template in templates:
            db_file = template.file.strip("\n")
            db_args = template.define_args.splitlines()

            include_list = []
            for arg in template.include_args:
                if arg in instance.__dict__:
                    include_list.append(f"{arg}={{{{ {arg} }}}}")
                else:
                    raise ValueError(
                        f"include arg '{arg}' in database template "
                        f"'{template.file}' not found in context"
                    )

            db_arg_string = ", ".join(db_args + include_list)

            jinja_txt += (
                f'msi -I${{EPICS_DB_INCLUDE_PATH}} -M"{db_arg_string}" "{db_file}"\n'
            )

        jinja_template = Template(jinja_txt)
        db_txt = jinja_template.render(instance.__dict__)  # type: ignore

        return db_txt + "\n"

    def render_environment_variables(self, instance: Entity) -> Optional[str]:
        """
        render the environment variable elements by combining the jinja template
        from an entity with the arguments from an Entity
        """
        variables = getattr(instance.__definition__, "env_vars")
        if not variables:
            return None

        env_var_txt = ""
        for variable in variables:
            # Substitute the name and value of the environment variable from args
            env_template = Template(f"epicsEnvSet {variable.name} {variable.value}")
            env_var_txt += env_template.render(instance.__dict__)
        return env_var_txt + "\n"

    def render_post_ioc_init(self, instance: Entity) -> Optional[str]:
        """
        render the post-iocInit entries by combining the jinja template
        from an entity with the arguments from an Entity
        """
        script = ""
        for line in getattr(instance.__definition__, "post_ioc_init"):
            script += self.render_text(instance, line)

        return script

    def render_elements(
        self, ioc: IOC, method: Callable[[Entity], Union[str, None]]
    ) -> str:
        """
        Render elements of a given IOC instance based on calling the correct method
        """
        elements = ""
        for instance in ioc.entities:
            if instance.entity_enabled:
                element = method(instance)
                if element:
                    elements += element
        return elements

    def render_script_elements(self, ioc: IOC) -> str:
        """
        Render all of the startup script entries for a given IOC instance
        """
        return self.render_elements(ioc, self.render_script)

    def render_database_elements(self, ioc: IOC) -> str:
        """
        Render all of the DBLoadRecords entries for a given IOC instance
        """
        return self.render_elements(ioc, self.render_database)

    def render_environment_variable_elements(self, ioc: IOC) -> str:
        """
        Render all of the environment variable entries for a given IOC instance
        """
        return self.render_elements(ioc, self.render_environment_variables)

    def render_post_ioc_init_elements(self, ioc: IOC) -> str:
        """
        Render all of the post-iocInit elements for a given IOC instance
        """
        return self.render_elements(ioc, self.render_post_ioc_init)



from pathlib import Path

from typer.testing import CliRunner

from ibek.ioc import clear_entity_classes
from ibek.utils import UTILS

from .test_cli import run_cli

runner = CliRunner()


def test_example_ioc(tmp_path: Path, samples: Path, ibek_defs: Path):
    """
    build an ioc from yaml and verify the result

    NOTE: the system test in tests/sys-test.sh uses the same example-ibek-config
    but instead of verifying the output, it runs the ioc in a container and \
    verifies that it starts up correctly.
    """
    clear_entity_classes()
    UTILS.__reset__()

    tmp_path = Path("/tmp/ibek_test")
    tmp_path.mkdir(exist_ok=True)

    entity_file = samples / "example-ibek-config" / "ioc.yaml"
    definition_file = ibek_defs / "_global" / "epics.ibek.support.yaml"
    definition_file2 = ibek_defs / "_global" / "devIocStats.ibek.support.yaml"
    out_file = tmp_path / "new_dir" / "test.ioc.cmd"
    out_db = tmp_path / "new_dir" / "make_db.sh"

    run_cli(
        "build-startup",
        entity_file,
        definition_file,
        definition_file2,
        "--out",
        out_file,
        "--db-out",
        out_db,
    )

    example_boot = (samples / "boot_scripts" / "test.ioc.cmd").read_text()
    actual_boot = out_file.read_text()

    example_db = (samples / "boot_scripts" / "test.ioc.make_db.sh").read_text()
    actual_db = out_db.read_text()

    assert example_boot == actual_boot
    assert example_db == actual_db


def test_example_sr_rf_08(tmp_path: Path, samples: Path, ibek_defs: Path):
    """
    build IOC sr-rf-ioc-08 from yaml and verify the result. This IOC tests:

    - all the yaml files in ibek-defs are schema compliant
    - the Utils class counter function (for Vectors and Carriers)
    - the once field (for Hy8401 comments)
    """

    clear_entity_classes()
    UTILS.__reset__()

    tmp_path = Path("/tmp/ibek_test2")
    tmp_path.mkdir(exist_ok=True)

    entity_file = samples / "example-srrfioc08" / "SR-RF-IOC-08.ibek.ioc.yaml"
    definition_files = ibek_defs.glob("*/*.support.yaml")
    out_file = tmp_path / "new_dir" / "st.cmd"
    out_db = tmp_path / "new_dir" / "make_db.sh"

    params = [
        "build-startup",
        entity_file,
        "--out",
        out_file,
        "--db-out",
        out_db,
    ]
    params += definition_files

    run_cli(*params)

    example_boot = (samples / "example-srrfioc08" / "st.cmd").read_text()
    actual_boot = out_file.read_text()

    example_db = (samples / "example-srrfioc08" / "make_db.sh").read_text()
    actual_db = out_db.read_text()

    assert example_boot == actual_boot
    assert example_db == actual_db


def test_values_ioc(tmp_path: Path, samples: Path, ibek_defs: Path):
    """
    build values ioc from yaml and verify the result

    This IOC verifies that repeated reference to a 'values' field
    with counter gets the same value every time.

    TODO: IMPORTANT: this test currently proves that the values are NOT the same
    TODO: make sure samples/values_test/st.cmd is updated when this is fixed.
    """

    clear_entity_classes()
    UTILS.__reset__()

    tmp_path = Path("/tmp/ibek_test2")
    tmp_path.mkdir(exist_ok=True)
    test_path = samples / "values_test"

    entity_file = test_path / "values.ibek.ioc.yaml"
    definition_files = test_path.glob("*.support.yaml")
    out_file = tmp_path / "new_dir" / "st.cmd"
    out_db = tmp_path / "new_dir" / "make_db.sh"

    params = [
        "build-startup",
        entity_file,
        "--out",
        out_file,
        "--db-out",
        out_db,
    ]
    params += definition_files

    run_cli(*params)

    example_boot = (test_path / "st.cmd").read_text()
    actual_boot = out_file.read_text()
    assert example_boot == actual_boot



"""
Tests for the rendering of scripts and database entries from generated
Entity classes
"""
from unittest.mock import Mock

from ibek.ioc import IOC
from ibek.render import Render


def test_pmac_asyn_ip_port_script(pmac_classes):
    generated_class = pmac_classes.PmacAsynIPPort
    pmac_asyn_ip = generated_class(name="my_pmac_instance", IP="111.111.111.111")

    render = Render()
    script_txt = render.render_script(pmac_asyn_ip)
    assert script_txt == "pmacAsynIPConfigure(my_pmac_instance, 111.111.111.111:1025)\n"


def test_geobrick_script(pmac_classes):
    generated_class = pmac_classes.Geobrick
    ip_port = Mock()
    ip_port.name = "my_asyn_port"
    pmac_geobrick_instance = generated_class(
        name="test_geobrick",
        PORT=ip_port,
        P="geobrick_one",
        numAxes=8,
        idlePoll=200,
        movingPoll=800,
    )

    render = Render()
    script_txt = render.render_script(pmac_geobrick_instance)

    assert (
        script_txt
        == "pmacCreateController(test_geobrick, my_asyn_port, 0, 8, 800, 200)\n"
        "pmacCreateAxes(test_geobrick, 8)\n"
    )


def test_geobrick_database(pmac_classes):
    generated_class = pmac_classes.Geobrick
    pmac_geobrick_instance = generated_class(
        name="test_geobrick",
        PORT="my_asyn_port",
        P="geobrick_one",
        numAxes=8,
        idlePoll=200,
        movingPoll=800,
    )

    render = Render()
    db_txt = render.render_database(pmac_geobrick_instance)

    assert (
        db_txt == 'msi -I${EPICS_DB_INCLUDE_PATH} -M"NAXES=8, PORT=test_geobrick, '
        "P=geobrick_one, idlePoll=200, movingPoll=800, TIMEOUT=4, CSG0=, CSG1=, "
        'CSG2=, CSG3=, CSG4=" "pmacController.template"\n'
        'msi -I${EPICS_DB_INCLUDE_PATH} -M"PORT=test_geobrick, '
        'P=geobrick_one" "pmacStatus.template"\n'
    )


def test_epics_environment_variables(epics_classes):
    # Using a specific variable entity
    generated_class = epics_classes.EpicsCaMaxArrayBytes
    max_array_bytes_instance = generated_class(max_bytes=10000000)

    render = Render()
    env_text = render.render_environment_variables(max_array_bytes_instance)

    assert env_text == "epicsEnvSet EPICS_CA_MAX_ARRAY_BYTES 10000000\n"

    # Using the generic entity
    env_name = "EPICS_CA_SERVER_PORT"
    env_value = 6000
    generated_class = epics_classes.EpicsEnvSet
    epics_env_set_instance = generated_class(name=env_name, value=env_value)

    env_text = render.render_environment_variables(epics_env_set_instance)

    assert env_text == f"epicsEnvSet {env_name} {env_value}\n"


def test_entity_disabled_does_not_render_elements(pmac_classes, epics_classes):
    # There are four elements to check: script, database, environment variables
    # and post-iocInit

    # Entity which has a script and database
    pmac_geobrick_class = pmac_classes.Geobrick
    # Entity which has env_vars
    ca_max_array_bytes_class = epics_classes.EpicsCaMaxArrayBytes
    # Entity which has post_ioc_init
    dbpf_class = epics_classes.Dbpf

    # We require pmac asyn IP port instances for the Geobrick class
    pmac_asyn_ip_class = pmac_classes.PmacAsynIPPort
    brick_one_asyn_ip_port = pmac_asyn_ip_class(
        name="geobrick_one_port", IP="172.23.100.101"
    )
    brick_two_asyn_ip_port = pmac_asyn_ip_class(
        name="geobrick_two_port", IP="172.23.100.101"
    )

    # Store created instances in a list
    instance_list = []

    # Create two instances of the Geobrick entity, one disabled
    instance_list.append(
        pmac_geobrick_class(
            name="geobrick_enabled",
            PORT=brick_one_asyn_ip_port,
            P="geobrick_one",
            numAxes=8,
            idlePoll=200,
            movingPoll=800,
        )
    )
    instance_list.append(
        pmac_geobrick_class(
            name="geobrick_disabled",
            PORT=brick_two_asyn_ip_port,
            P="geobrick_two",
            numAxes=8,
            idlePoll=200,
            movingPoll=800,
            entity_enabled=False,
        )
    )

    # Create two instances of the CA max array bytes entity, one disabled
    instance_list.append(ca_max_array_bytes_class())
    instance_list.append(ca_max_array_bytes_class(entity_enabled=False))

    # Create two instances of the dpbf class
    instance_list.append(dbpf_class(pv="TEST:PV:1", value="pv_value_1"))
    instance_list.append(
        dbpf_class(pv="TEST:PV:2", value="pv_value_2", entity_enabled=False)
    )

    # Make an IOC with our instances
    ioc = IOC(
        "TEST-IOC-01",
        "Test IOC",
        instance_list,
        "test_ioc_image",
    )

    # Render script and check output
    expected_script = (
        "pmacCreateController(geobrick_enabled, geobrick_one_port, 0, 8, 800, 200)\n"
        "pmacCreateAxes(geobrick_enabled, 8)\n"
    )
    render = Render()
    script = render.render_script_elements(ioc)
    assert script == expected_script

    # Render database
    expected_database = (
        'msi -I${EPICS_DB_INCLUDE_PATH} -M"NAXES=8, PORT=geobrick_enabled, '
        "P=geobrick_one, idlePoll=200, movingPoll=800, TIMEOUT=4, CSG0=, CSG1=, "
        'CSG2=, CSG3=, CSG4=" "pmacController.template"\n'
        'msi -I${EPICS_DB_INCLUDE_PATH} -M"PORT=geobrick_enabled, P=geobrick_one" '
        '"pmacStatus.template"\n'
    )
    database = render.render_database_elements(ioc)
    assert database == expected_database

    # Render environment variables
    expected_env_vars = "epicsEnvSet EPICS_CA_MAX_ARRAY_BYTES 6000000\n"
    env_vars = render.render_environment_variable_elements(ioc)
    assert env_vars == expected_env_vars

    # Render post_ioc_init
    expected_post_ioc_init = 'dbpf "TEST:PV:1", "pv_value_1"\n'
    post_ioc_init = render.render_post_ioc_init_elements(ioc)
    assert post_ioc_init == expected_post_ioc_init



import json
import subprocess
import sys
from pathlib import Path

import pytest
from ruamel.yaml import YAML
from typer.testing import CliRunner

from ibek import __version__
from ibek.__main__ import cli
from ibek.ioc import clear_entity_classes, make_entity_classes
from ibek.support import Support

runner = CliRunner()


def test_cli_version():
    cmd = [sys.executable, "-m", "ibek", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__


def run_cli(*args):
    result = runner.invoke(cli, [str(x) for x in args])
    if result.exception:
        raise result.exception
    assert result.exit_code == 0, result


def test_builder_schema(tmp_path: Path, samples: Path):
    """generate the global ibek schema"""
    schema_path = tmp_path / "schema.json"
    run_cli("ibek-schema", schema_path)
    expected = json.loads((samples / "schemas" / "ibek.defs.schema.json").read_text())
    # Don't care if version number didn't update to match if the rest is the same
    # expected["title"] = mock.ANY

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_pmac_schema(tmp_path: Path, samples: Path, ibek_defs: Path):
    """generate schema from the pmac support module definition yaml"""
    clear_entity_classes()

    schema_path = tmp_path / "pmac.ibek.entities.schema.json"
    yaml_path = ibek_defs / "pmac" / "pmac.ibek.support.yaml"
    run_cli("ioc-schema", yaml_path, schema_path)

    expected = json.loads(
        (samples / "schemas" / "pmac.ibek.entities.schema.json").read_text()
    )

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_asyn_schema(tmp_path: Path, samples: Path, ibek_defs: Path):
    """generate schema from the asyn support module definition yaml"""
    clear_entity_classes()

    schema_path = tmp_path / "asyn.ibek.entities.schema.json"
    yaml_path = ibek_defs / "asyn" / "asyn.ibek.support.yaml"
    run_cli("ioc-schema", yaml_path, schema_path)

    expected = json.loads(
        (samples / "schemas" / "asyn.ibek.entities.schema.json").read_text()
    )

    actual = json.loads((schema_path).read_text())
    assert expected == actual


def test_container_schema(tmp_path: Path, samples: Path, ibek_defs: Path):
    clear_entity_classes()
    """generate schema for a container with two support modules"""

    schema_combined = tmp_path / "container.ibek.entities.schema.json"
    yaml_path1 = ibek_defs / "asyn" / "asyn.ibek.support.yaml"
    yaml_path2 = ibek_defs / "pmac" / "pmac.ibek.support.yaml"
    run_cli("ioc-schema", yaml_path1, yaml_path2, schema_combined)

    expected = json.loads(
        (samples / "schemas" / "container.ibek.entities.schema.json").read_text()
    )

    actual = json.loads((schema_combined).read_text())
    assert expected == actual


def test_build_startup_output_path(tmp_path: Path, samples: Path, ibek_defs: Path):
    """
    build an ioc startup script and ensure output directory gets generated
    if it doesn't pre-exist
    """
    clear_entity_classes()
    entity_file = samples / "yaml" / "bl45p-mo-ioc-02.ibek.ioc.yaml"
    definition_file = ibek_defs / "pmac" / "pmac.ibek.support.yaml"
    out_file = tmp_path / "new_dir" / "st.cmd"
    out_db = tmp_path / "new_dir" / "make_db.sh"

    run_cli(
        "build-startup",
        entity_file,
        definition_file,
        "--out",
        out_file,
        "--db-out",
        out_db,
    )

    example_boot = (samples / "boot_scripts" / "st.cmd").read_text()
    actual_boot = out_file.read_text()

    assert example_boot == actual_boot


def test_build_startup_single(tmp_path: Path, samples: Path, ibek_defs: Path):
    """
    build an ioc startup script from an IOC instance entity file
    and a single support module definition file
    """
    clear_entity_classes()
    entity_file = samples / "yaml" / "bl45p-mo-ioc-02.ibek.ioc.yaml"
    definition_file = ibek_defs / "pmac" / "pmac.ibek.support.yaml"
    out_file = tmp_path / "st.cmd"
    out_db = tmp_path / "make_db.sh"

    run_cli(
        "build-startup",
        entity_file,
        definition_file,
        "--out",
        out_file,
        "--db-out",
        out_db,
    )

    example_boot = (samples / "boot_scripts" / "st.cmd").read_text()
    actual_boot = out_file.read_text()

    assert example_boot == actual_boot


def test_build_startup_multiple(tmp_path: Path, samples: Path, ibek_defs: Path):
    """
    build an ioc startup script from an IOC instance entity file
    and multiple support module definition files
    """
    clear_entity_classes()
    entity_file = samples / "yaml" / "bl45p-mo-ioc-03.ibek.ioc.yaml"
    definition_file1 = ibek_defs / "asyn" / "asyn.ibek.support.yaml"
    definition_file2 = ibek_defs / "pmac" / "pmac.ibek.support.yaml"
    out_file = tmp_path / "st.cmd"
    out_db = tmp_path / "make_db.sh"

    run_cli(
        "build-startup",
        entity_file,
        definition_file1,
        definition_file2,
        "--out",
        out_file,
        "--db-out",
        out_db,
    )

    example_boot = (samples / "boot_scripts" / "stbl45p-mo-ioc-03").read_text()
    actual_boot = out_file.read_text()

    assert example_boot == actual_boot


def test_build_startup_env_vars_and_post_ioc_init(
    tmp_path: Path, samples: Path, ibek_defs: Path
):
    """
    build an ioc startup script from an IOC instance entity file, multiple
    support module definition files which include environment variables and
    post iocInit() entries
    """
    clear_entity_classes()
    entity_file = samples / "yaml" / "bl45p-mo-ioc-04.ibek.ioc.yaml"
    definition_file1 = ibek_defs / "_global" / "epics.ibek.support.yaml"
    definition_file2 = ibek_defs / "pmac" / "pmac.ibek.support.yaml"
    out_file = tmp_path / "st.cmd"
    out_db = tmp_path / "make_db.sh"

    run_cli(
        "build-startup",
        entity_file,
        definition_file1,
        definition_file2,
        "--out",
        out_file,
        "--db-out",
        out_db,
    )

    example_boot = (samples / "boot_scripts" / "stbl45p-mo-ioc-04").read_text()
    actual_boot = out_file.read_text()

    assert example_boot == actual_boot


def test_loading_module_twice(tmp_path: Path, samples: Path, ibek_defs: Path):
    """
    regression test to demonstrate that clear_entity_classes works and
    allows us to call make_entity_classes more than once
    """

    clear_entity_classes()
    # When we deserialize the same yaml twice as we do in the full test suite
    # we may get clashes in the namespace of generated Entity classes.
    # This tests that we get a sensible error when we do
    definition_file = ibek_defs / "pmac" / "pmac.ibek.support.yaml"
    support = Support.deserialize(YAML(typ="safe").load(definition_file))
    make_entity_classes(support)
    with pytest.raises(AssertionError) as ctx:
        make_entity_classes(support)
    assert str(ctx.value) == "Entity classes already created for pmac"


def test_bad_counter(tmp_path: Path, samples: Path):
    """
    Check you cannot redefine a counter with the same name and different params
    """

    clear_entity_classes()
    entity_file = samples / "yaml" / "bad_counter.ibek.ioc.yaml"
    definition_file1 = samples / "yaml" / "bad_counter.ibek.support.yaml"

    with pytest.raises(ValueError) as ctx:
        run_cli("build-startup", entity_file, definition_file1)
    assert (
        str(ctx.value)
        == "Redefining counter InterruptVector with different start/stop values"
    )



from ibek.ioc import IOC, clear_entity_classes, id_to_entity, make_entity_classes
from ibek.support import Definition, IdArg, ObjectArg, Support


def test_conversion_classes():
    clear_entity_classes()
    support = Support(
        "mymodule",
        [
            Definition("port", "A port", [IdArg("name", "the name", "id")]),
            Definition("device", "a device", [ObjectArg("port", "the port", "object")]),
        ],
    )
    namespace = make_entity_classes(support)
    assert {"device", "port"}.issubset(dir(namespace))
    assert namespace.port.__definition__ == support.defs[0]
    assert namespace.device.__definition__ == support.defs[1]
    d = dict(
        ioc_name="",
        description="",
        entities=[
            dict(type="mymodule.port", name="PORT"),
            dict(type="mymodule.device", port="PORT"),
        ],
        generic_ioc_image="",
    )
    ioc = IOC.deserialize(d)
    port, device = ioc.entities
    assert port.type == "mymodule.port"
    assert id_to_entity == {"PORT": port}
    assert device.type == "mymodule.device"
    assert device.port is port



from pathlib import Path

from typer.testing import CliRunner

from ibek.__main__ import cli
from ibek.ioc import clear_entity_classes

runner = CliRunner()


def run_cli(*args):
    result = runner.invoke(cli, [str(x) for x in args])
    if result.exception:
        raise result.exception
    assert result.exit_code == 0, result


"""
A test for debugging the creation of startup scripts
remove xx to run this in debug mode.
"""


def xxtest_build_startup_p45(tmp_path: Path, samples: Path):
    """
    build an ioc startup script from an IOC instance entity file
    and multiple support module definition files
    """
    root = Path("/workspace/bl45p")

    clear_entity_classes()
    entity_file = root / "bl45p-mo-ioc-99.yaml"
    definitions = Path.glob(root / "ibek", "*.ibek.support.yaml")
    out_file = root / "iocs/bl45p-mo-ioc-99/config/st.cmd"

    run_cli("build-startup", entity_file, *definitions, "--out", out_file)



"""
Functions for building the db and boot scripts
"""
import logging
import re
from pathlib import Path
from typing import List

from jinja2 import Template
from ruamel.yaml.main import YAML

from .ioc import IOC, make_entity_classes
from .render import Render
from .support import Support

log = logging.getLogger(__name__)

TEMPLATES = Path(__file__).parent / "templates"

schema_modeline = re.compile(r"# *yaml-language-server *: *\$schema=([^ ]*)")
url_f = r"file://"


def ioc_deserialize(ioc_instance_yaml: Path, definition_yaml: List[Path]) -> IOC:
    """
    Takes an ioc instance entities file, list of generic ioc definitions files.

    Returns an in memory object graph of the resulting ioc instance
    """

    # Read and load the support module definitions
    for yaml in definition_yaml:
        support = Support.deserialize(YAML(typ="safe").load(yaml))
        make_entity_classes(support)

    # Create an IOC instance from it
    return IOC.deserialize(YAML(typ="safe").load(ioc_instance_yaml))


def create_db_script(ioc_instance: IOC) -> str:
    """
    Create make_db.sh script for expanding the database templates
    """
    with open(TEMPLATES / "make_db.jinja", "r") as f:
        template = Template(f.read())

    renderer = Render()

    return template.render(
        database_elements=renderer.render_database_elements(ioc_instance),
    )


def create_boot_script(ioc_instance: IOC) -> str:
    """
    Create the boot script for an IOC
    """
    with open(TEMPLATES / "st.cmd.jinja", "r") as f:
        template = Template(f.read())

    renderer = Render()

    return template.render(
        env_var_elements=renderer.render_environment_variable_elements(ioc_instance),
        script_elements=renderer.render_script_elements(ioc_instance),
        post_ioc_init_elements=renderer.render_post_ioc_init_elements(ioc_instance),
    )
